package org.embulk.input.filename;

import java.util.List;
import java.util.ArrayList;
import java.util.Collections;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.Files;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.FileVisitResult;
import java.nio.file.attribute.BasicFileAttributes;
import com.google.common.collect.ImmutableList;
import com.google.common.base.Optional;
import org.slf4j.Logger;
import org.embulk.config.Config;
import org.embulk.config.ConfigDefault;
import org.embulk.config.ConfigInject;
import org.embulk.config.ConfigSource;
import org.embulk.config.ConfigDiff;
import org.embulk.config.TaskReport;
import org.embulk.config.Task;
import org.embulk.config.TaskSource;
import org.embulk.spi.Exec;
import org.embulk.spi.FileInputPlugin;
import org.embulk.spi.BufferAllocator;
import org.embulk.spi.TransactionalFileInput;
import org.embulk.spi.util.InputStreamTransactionalFileInput;
import org.embulk.standards.LocalFileInputPlugin;

import java.nio.file.attribute.BasicFileAttributeView;
import java.nio.file.attribute.FileTime;
import java.util.Comparator;



public class FilenameFileInputPlugin implements FileInputPlugin
{

    public interface PluginTask extends Task
    {
        @Config("multi_dir")
        @ConfigDefault("[]")
        List<String> getMultiDir();
        
        @Config("multi_tag")
        @ConfigDefault("[]")
        List<String> getMultiTag();
        
        
        @Config("path_prefix")
        @ConfigDefault("")
        String getPathPrefix();

        @Config("last_path")
        @ConfigDefault("null")
        Optional<String> getLastPath();
        
        @Config("order_by_modified_time")
        @ConfigDefault("0")
        int getOrderByModifiedTime();
        
        @Config("order_by_creation_time")
        @ConfigDefault("0")
        int getOrderByCreationTime();

        @Config("file_size")
        @ConfigDefault("null")
        Optional<Integer> getFileSize();

        @Config("follow_symlinks")
        @ConfigDefault("false")
        boolean getFollowSymlinks();

        List<String> getFiles();
        void setFiles(List<String> files);

        @ConfigInject
        BufferAllocator getBufferAllocator();
    }

    public static FileTime getCreationTime(String filename) throws IOException{
      File file = new File(filename);
      Path p = Paths.get(file.getAbsolutePath());
      BasicFileAttributes view = Files.getFileAttributeView(p,BasicFileAttributeView.class).readAttributes();
      FileTime fileTime = view.creationTime();
      //System.out.println("The raw creation time of " +filename+  " is " + fileTime.toString());
      return fileTime;
    }
  
    public static FileTime getLastModifiedTime(String filename) throws IOException{
      File file = new File(filename);
      Path p = Paths.get(file.getAbsolutePath());
      BasicFileAttributes view = Files.getFileAttributeView(p,BasicFileAttributeView.class).readAttributes();
      FileTime fileTime = view.lastModifiedTime();
      //System.out.println("The raw last modified time of " +filename+  " is " + fileTime.toString());
      return fileTime;
    }

    private final Logger log = Exec.getLogger(getClass());

    private final static Path CURRENT_DIR = Paths.get(".").normalize();
    
    public static String theTag = "";
    
    public static List<Integer> tagIndex = new ArrayList<Integer>();
    
    public static List<String> tagList;
    

    @Override
    public ConfigDiff transaction(ConfigSource config, FileInputPlugin.Control control)
    {
        
        PluginTask task = config.loadConfig(PluginTask.class);
        
        List<String> allFiles = new ArrayList<String> ();
        
        tagIndex.add(0);
        //int s = 0;
        
        List<String> dirList = task.getMultiDir();
        tagList = task.getMultiTag();
        if ( dirList.size() != 0 ) {
            log.info("The list of dir: " + dirList);
            while (tagList.size() < dirList.size()){
                tagList.add("");
            }
        } else {
            if (task.getPathPrefix().equals("")){
                throw new RuntimeException("Please input the path_prefix or the multi_dir");
            }
            dirList.add(task.getPathPrefix());
            log.info("list of dir: " + dirList);
            tagList.add("");
        }
        

        // list files recursively
        ConfigDiff res = Exec.newConfigDiff();
        for (int i=0; i< dirList.size();i++)
        {
            flag = 0;
            List<String> files = listFiles(task,Paths.get(dirList.get(i)).normalize());
            
            //Sort the listFiles according to the configuration.
            int order_modified = task.getOrderByModifiedTime();
            int order_creation = task.getOrderByCreationTime();
        
            if (order_modified == 0 && order_creation == 0){
                Collections.sort(files);
            } else if(order_creation == 0){
                Collections.sort(files,new Comparator<String>(){
                    @Override
                    public int compare(String f1, String f2) {
                        try{
                            return getLastModifiedTime(f1).compareTo(getLastModifiedTime(f2));
                        } catch (IOException ex){
                            ex.printStackTrace();
                        }
                        return 0;
                    }
                });
          
                if (order_modified == 1 ) { Collections.reverse(files); }
            
            } else if (order_modified == 0 ){
                Collections.sort(files,new Comparator<String>(){
                    @Override
                    public int compare(String f1, String f2) {
                        try{
                            return getCreationTime(f1).compareTo(getCreationTime(f2));
                        } catch (IOException ex){
                            ex.printStackTrace();
                        }
                        return 0;
                    }
                });
          
                if ( order_creation == 1 ) { Collections.reverse(files);}
            } else {
                throw new RuntimeException("Could not order by creation time and lasModified time at the same time");
            }
        
            log.info("Loading files {}", files);
            allFiles.addAll(files);
            //task.setFiles(files);
        
            //s += files.size()
            tagIndex.add(allFiles.size());
            //taskList.add(task.deepCopy);
        // number of processors is same with number of files
        
            //int taskCount = files.size();
            //theTag = tagList.get(i);
            //info.log();
            //res = resume(task.dump(), taskCount, control);
        }
        
        task.setFiles(allFiles);
        
        int taskCount = allFiles.size();
        //return res;
        return resume(task.dump(), taskCount, control);
    }

    @Override
    public ConfigDiff resume(TaskSource taskSource,
            int taskCount,
            FileInputPlugin.Control control)
    {
        PluginTask task = taskSource.loadTask(PluginTask.class);
        
        // Here the taskSource contains all the Configuration of the 'in'
        log.info("The taskSource of the FileName in the ConfigDiff resume: " + taskSource.toString());
        
        // Here will run all the tasks. Each task is to deal with a file.
        control.run(taskSource, taskCount);
        
        // build next config
        ConfigDiff configDiff = Exec.newConfigDiff();
        
        // last_path
        if (task.getFiles().isEmpty()) {
            // keep the last value
            if (task.getLastPath().isPresent()) {
                configDiff.set("last_path", task.getLastPath().get());
            }
        } else {
            List<String> files = new ArrayList<String>(task.getFiles());
            log.info("The File order is {}",files);
            configDiff.set("last_path", files.get(files.size() - 1));
        }
        return configDiff;
    }

    @Override
    public void cleanup(TaskSource taskSource,
            int taskCount,
            List<TaskReport> successTaskReports)
    { }

    public List<String> listFiles(PluginTask task,Path pathPrefix)
    {
        //Path pathPrefix = Paths.get(task.getPathPrefix()).normalize();
        final Path directory;
        final String fileNamePrefix;
        if (Files.isDirectory(pathPrefix)) {
            directory = pathPrefix;
            fileNamePrefix = "";
        } else {
            fileNamePrefix = pathPrefix.getFileName().toString();
            Path d = pathPrefix.getParent();
            directory = (d == null ? CURRENT_DIR : d);
        }

        //final ImmutableList.Builder<String> builder = ImmutableList.builder();
        final List<String> filesArray = new ArrayList<String>();
        final String lastPath = task.getLastPath().orNull();
        final Integer fileSize = task.getFileSize().orNull();
        try {
            log.info("Listing local files at directory '{}' filtering filename by prefix '{}'", directory.equals(CURRENT_DIR) ? "." : directory.toString(), fileNamePrefix);
            Files.walkFileTree(directory, new SimpleFileVisitor<Path>() {
                @Override
                public FileVisitResult preVisitDirectory(Path path, BasicFileAttributes attrs)
                {
                    if (path.equals(directory)) {
                        return FileVisitResult.CONTINUE;
                    } else if (lastPath != null && path.toString().compareTo(lastPath.substring(0, path.toString().length())) < 0) {
                        return FileVisitResult.SKIP_SUBTREE;
                    } else if (path.getFileName().toString().startsWith(".")) {
                        return FileVisitResult.SKIP_SUBTREE;
                    } else {
                        if (path.getFileName().toString().startsWith(fileNamePrefix)) {
                            return FileVisitResult.CONTINUE;
                        } else {
                            return FileVisitResult.SKIP_SUBTREE;
                        }
                    }
                }


                @Override
                public FileVisitResult visitFile(Path path, BasicFileAttributes attrs)
                {
                    if (lastPath != null && path.toString().compareTo(lastPath) <= 0) {
                        return FileVisitResult.CONTINUE;
                    } else if (path.getFileName().toString().startsWith(".")) {
                         return FileVisitResult.CONTINUE;
                    } else {
                        if (path.getFileName().toString().startsWith(fileNamePrefix)) {
                            if (fileSize == null || path.toFile().length() == fileSize) {
                              //builder.add(path.toString());
                              filesArray.add(path.toString());
                            }
                        }
                        return FileVisitResult.CONTINUE;
                    }
               }
            });
        } catch (IOException ex) {
            throw new RuntimeException(String.format("Failed get a list of local files at '%s'", directory), ex);
        }
        //return builder.build();
        return filesArray;
    }

    @Override
    public TransactionalFileInput open(TaskSource taskSource, int taskIndex)
    {
        final PluginTask task = taskSource.loadTask(PluginTask.class);
        
        log.info("The task in open: " + taskSource.toString());
        log.info("The taskIndex: " + taskIndex);
        final String path = task.getFiles().get(taskIndex);
        
        setTag(taskIndex);
        log.info("The tag: " + theTag);

        return new InputStreamTransactionalFileInput(
                task.getBufferAllocator(),
                new InputStreamTransactionalFileInput.Opener() {
                    public InputStream open() throws IOException
                    {
                        return new FilenameFileInputStream(path);
                    }
                })
        {
            @Override
            public void abort()
            { }

            @Override
            public TaskReport commit()
            {
                return Exec.newTaskReport();
            }
        };
    }
    
    
    public static int flag = 0;
    
    public static void setTag(int index)
    {
        if (index == tagIndex.get(flag))
        {
            flag+=1;
        }
        theTag = tagList.get(flag-1);
    }

    class FilenameFileInputStream extends FileInputStream {
        final int MAX_NAME_LENGTH = 255;
        int n;
        byte[] name;
    
        FilenameFileInputStream(File file) throws FileNotFoundException {
            super(file);
            n = 0;
            name = (theTag+file.getName()).getBytes();
        }
    
        FilenameFileInputStream(String path) throws FileNotFoundException {
            super(path);
            n = 0;
            name = (theTag+path).getBytes();
        }
    
        @Override
        public int read() throws IOException {
            if (n < name.length) {
                byte b = name[n];
                n++;
                return b;
            } else if (n < MAX_NAME_LENGTH) {
                n++;
                return 0;
            } else {
                return super.read();
            }
        }
    
        @Override
        public int read(byte[] b) throws IOException {
            return read(b, 0, b.length);
        }
    
        @Override
        public int read(byte[] b, int off, int len) throws IOException {
            if (n < MAX_NAME_LENGTH) {
                int i = 0;
                int c;
                for (; i < len; i++) {
                    c = read();
                    if (c == -1) {
                        if ( i == 0 ) {
                            return -1;
                        }
                        break;
                    }
                    b[off + i] = (byte)c;
                }
                return i;
             } else {
                return super.read(b, off, len);
             }
        }
    }
}