package org.embulk.input.filename;

import org.embulk.config.Config;
import org.embulk.config.ConfigDefault;
import org.embulk.config.ConfigDiff;
import org.embulk.config.ConfigSource;
import org.embulk.config.Task;
import org.embulk.config.TaskSource;
import org.embulk.spi.ParserPlugin;
import org.embulk.spi.FileInput;
import org.embulk.spi.PageOutput;
import org.embulk.spi.Schema;
import org.embulk.spi.SchemaConfig;

import org.embulk.spi.Exec;
import org.embulk.spi.PageBuilder;
import org.embulk.spi.util.FileInputInputStream;
import org.embulk.spi.ColumnConfig;
import java.io.IOException;
import java.util.Arrays;
import java.util.ArrayList;
import org.apache.commons.codec.binary.Base64;

import static org.embulk.spi.type.Types.STRING;

import org.slf4j.Logger;


public class NoneBinParserPlugin
        implements ParserPlugin
{
    static int MAX_NAME_LENGTH = 255;
    Schema schema;

    public interface PluginTask
            extends Task //, LineDecoder.DecoderTask //, TimestampParser.Task
    {
        @Config("column_name")
        @ConfigDefault("\"payload\"")
        public String getColumnName();
    }

    private final Logger log;

    public NoneBinParserPlugin()
    {
        this.log = Exec.getLogger(NoneBinParserPlugin.class);
    }

    @Override
    public void transaction(ConfigSource config, ParserPlugin.Control control)
    {
        PluginTask task = config.loadConfig(PluginTask.class);
        log.info("The ConfigSource is: " + config.toString());
        ArrayList<ColumnConfig> columns = new ArrayList<ColumnConfig>();
        final String columnName = task.getColumnName();

        columns.add(new ColumnConfig(columnName, STRING, config));
        columns.add(new ColumnConfig("tag", STRING, config)); 

        // In the Unit test we need to convert the  output of the parser to java object 
        // Such conversion is based on the parser's schema so that we need keep this schema in parser instance's variable instead
        // of using it just once in this method.
        this.schema = new SchemaConfig(columns).toSchema();
        control.run(task.dump(), this.schema);
    }

    @Override
    public void run(TaskSource taskSource, Schema schema, 
            FileInput input, PageOutput output)
    {
        PluginTask task = taskSource.loadTask(PluginTask.class);
        log.info("The taskSource of the Parser: "+ taskSource.toString());
        FileInputInputStream dataIn = new FileInputInputStream(input);
        PageBuilder pageBuilder = new PageBuilder(Exec.getBufferAllocator(), schema, output);
        int chunksize = 1024 * 1024 * 1;
        while( input.nextFile() ){
            byte[] pathBytesArray = new byte[MAX_NAME_LENGTH];
            int i = 0;
            int c;
            for (; i < MAX_NAME_LENGTH; i++) {
                c = dataIn.read();
                if ( c == -1) {
                    break;
                } else if ( c == 0 ) {
                    // read empty bytes until MAX_NAME_LENGTH;
                    for (int j = i + 1; j < MAX_NAME_LENGTH; j++) {
                        dataIn.read();
                    }
                    break;
                }
                pathBytesArray[i] = (byte)c;
            }
            String path = new String(Arrays.copyOfRange(pathBytesArray, 0, i));

            // To read the data, we read one byte from the dataIn, if it isn't the end of file we record it to the bytesArray, 
            // we jugde the length of the added bytes, if len == chunksize we record bytesArray to the page record the bytesArray again 
            int bytes_read = 0;
            bytes_read = dataIn.read();
            int len = 0;
            byte[] bytesArray = new byte[chunksize];
            while(bytes_read != -1) {
                // Read one byte from the dataIn and record it to the bytesArray
                bytesArray[len] = (byte) bytes_read;
                bytes_read = dataIn.read();
                len += 1 ;
                if (len == chunksize) {
                    log.info(path);
                    pageBuilder.setString(0, Base64.encodeBase64String(bytesArray));
                    pageBuilder.setString(1, path);
                    pageBuilder.addRecord();
                    len = 0;
               }
            }

            // In case the the remain part of the data is less than chunksize we need to record it to the page as well.
            if (len != 0) {
               pageBuilder.setString(0,Base64.encodeBase64String(Arrays.copyOfRange(bytesArray, 0, len)));
               pageBuilder.setString(1,path);
               pageBuilder.addRecord();
            }
        }
        pageBuilder.finish();
    }
}
